@@ -13,6 +13,7 @@ import argparse
 import atexit
 import hashlib
 import json
+import logging
 import os.path
 import platform
 import random
@@ -26,6 +27,13 @@ import threading
 import pytest
 
 from b2sdk.v1 import B2Api, InMemoryAccountInfo, InMemoryCache, fix_windows_path_limit
+from b2sdk.v1 import EncryptionAlgorithm, EncryptionMode, EncryptionSetting
+
+SSE_NONE = EncryptionSetting(mode=EncryptionMode.NONE,)
+SSE_B2_AES = EncryptionSetting(
+    mode=EncryptionMode.SSE_B2,
+    algorithm=EncryptionAlgorithm.AES256,
+)
 
 
 def parse_args(tests):
@@ -55,7 +63,7 @@ def parse_args(tests):
 
 def error_and_exit(message):
     print('ERROR:', message)
-    sys.exit(1)
+    _exit(1)
 
 
 def read_file(path):
@@ -169,7 +177,13 @@ def print_json_indented(value):
     """
     Converts the value to JSON, then prints it.
     """
-    print_text_indented(json.dumps(value, indent=4, sort_keys=True))
+    print_text_indented(json.dumps(value, indent=4, sort_keys=True, default=serialize_enc_settings))
+
+
+def serialize_enc_settings(value):
+    if not isinstance(value, EncryptionSetting):
+        raise TypeError
+    return value.as_value_dict()
 
 
 def print_output(status, stdout, stderr):
@@ -246,7 +260,7 @@ class CommandLine:
         status, stdout, stderr = run_command(self.command, args)
         if status != 0:
             print('FAILED with status', status)
-            sys.exit(1)
+            _exit(1)
         if stderr != '':
             failed = False
             for line in (s.strip() for s in stderr.split(os.linesep)):
@@ -256,7 +270,7 @@ class CommandLine:
             if failed:
                 print('FAILED because of stderr')
                 print(stderr)
-                sys.exit(1)
+                _exit(1)
         if expected_pattern is not None:
             if re.search(expected_pattern, stdout) is None:
                 print('STDOUT:')
@@ -280,7 +294,7 @@ class CommandLine:
         status, stdout, stderr = run_command(self.command, args)
         if status == 0:
             print('ERROR: should have failed')
-            sys.exit(1)
+            _exit(1)
         if re.search(expected_pattern, stdout + stderr) is None:
             print(expected_pattern)
             print(stdout + stderr)
@@ -302,10 +316,17 @@ def should_equal(expected, actual):
     print_json_indented(actual)
     if expected != actual:
         print('  ERROR')
-        sys.exit(1)
+        _exit(1)
     print()
 
 
+def _exit(error_code):
+    logging.shutdown()
+    sys.stdout.flush()
+    sys.stderr.flush()
+    sys.exit(error_code)
+
+
 def setup_envvar_test(envvar_name, envvar_value):
     """
     Establish config for environment variable test.
@@ -580,8 +601,33 @@ def file_version_summary(list_of_files):
     hide, looking like this:
 
        ['+ photos/a.jpg', '- photos/b.jpg', '+ photos/c.jpg']
+
     """
-    return [('+ ' if (f['action'] == 'upload') else '- ') + f['fileName'] for f in list_of_files]
+    return [filename_summary(f) for f in list_of_files]
+
+
+def filename_summary(file_):
+    return ('+ ' if (file_['action'] == 'upload') else '- ') + file_['fileName']
+
+
+def file_version_summary_with_encryption(list_of_files):
+    """
+    Given the result of list-file-versions, returns a list
+    of all file versions, with "+" for upload and "-" for
+    hide, with information about encryption, looking like this:
+
+       [
+           ('+ photos/a.jpg', 'SSE-B2:AES256'),
+           ('- photos/b.jpg', None),
+           ('+ photos/c.jpg', 'none'),
+       ]
+    """
+    result = []
+    for f in list_of_files:
+        entry = filename_summary(f)
+        encryption = encryption_summary(f['serverSideEncryption'])
+        result.append((entry, encryption))
+    return result
 
 
 def find_file_id(list_of_files, file_name):
@@ -591,15 +637,29 @@ def find_file_id(list_of_files, file_name):
     assert False, 'file not found: %s' % (file_name,)
 
 
+def encryption_summary(sse_dict):
+    if isinstance(sse_dict, EncryptionSetting):
+        sse_dict = sse_dict.as_value_dict()
+    encryption = sse_dict['mode']
+    algorithm = sse_dict.get('algorithm')
+    if algorithm is not None:
+        encryption += ':' + algorithm
+    return encryption
+
+
 def sync_up_test(b2_tool, bucket_name):
     sync_up_helper(b2_tool, bucket_name, 'sync')
 
 
+def sync_up_sse_b2_test(b2_tool, bucket_name):
+    sync_up_helper(b2_tool, bucket_name, 'sync', encryption=SSE_B2_AES)
+
+
 def sync_up_test_no_prefix(b2_tool, bucket_name):
     sync_up_helper(b2_tool, bucket_name, '')
 
 
-def sync_up_helper(b2_tool, bucket_name, dir_):
+def sync_up_helper(b2_tool, bucket_name, dir_, encryption=None):
     sync_point_parts = [bucket_name]
     if dir_:
         sync_point_parts.append(dir_)
@@ -627,18 +687,32 @@ def sync_up_helper(b2_tool, bucket_name, dir_):
         os.symlink('broken', p('d'))
 
         # now upload
-        b2_tool.should_succeed(
-            ['sync', '--noProgress', dir_path, b2_sync_point],
-            expected_pattern="d could not be accessed"
-        )
+        if encryption is None:
+            command = ['sync', '--noProgress', dir_path, b2_sync_point]
+            expected_encryption = SSE_NONE
+        elif encryption == SSE_B2_AES:
+            command = [
+                'sync', '--noProgress', '--destinationServerSideEncryption', 'SSE-B2', dir_path,
+                b2_sync_point
+            ]
+            expected_encryption = encryption
+        else:
+            raise ValueError('unsupported encryption mode: %s' % encryption)
+
+        b2_tool.should_succeed(command, expected_pattern="d could not be accessed")
         file_versions = b2_tool.list_file_versions(bucket_name)
+
+        expected_encryption_str = encryption_summary(expected_encryption.as_value_dict())
         should_equal(
             [
-                '+ ' + prefix + 'a',
-                '+ ' + prefix + 'b',
-                '+ ' + prefix + 'c',
-            ], file_version_summary(file_versions)
+                ('+ ' + prefix + 'a', expected_encryption_str),
+                ('+ ' + prefix + 'b', expected_encryption_str),
+                ('+ ' + prefix + 'c', expected_encryption_str),
+            ],
+            file_version_summary_with_encryption(file_versions),
         )
+        if encryption is not None:
+            return  # that's enough, we've checked that encryption works, no need to repeat the whole sync suite
 
         c_id = find_file_id(file_versions, prefix + 'c')
         file_info = b2_tool.should_succeed_json(['get-file-info', c_id])['fileInfo']
@@ -839,11 +913,23 @@ def sync_copy_test(b2_tool, bucket_name):
     sync_copy_helper(b2_tool, bucket_name, 'sync')
 
 
-def sync_copy_test_no_prefix(b2_tool, bucket_name):
-    sync_copy_helper(b2_tool, bucket_name, '')
+def sync_copy_test_no_prefix_default_encryption(b2_tool, bucket_name):
+    sync_copy_helper(b2_tool, bucket_name, '', encryption=None, expected_encryption=SSE_NONE)
 
 
-def sync_copy_helper(b2_tool, bucket_name, folder_in_bucket):
+def sync_copy_test_no_prefix_no_encryption(b2_tool, bucket_name):
+    sync_copy_helper(b2_tool, bucket_name, '', encryption=SSE_NONE, expected_encryption=SSE_NONE)
+
+
+def sync_copy_test_no_prefix_sse_b2(b2_tool, bucket_name):
+    sync_copy_helper(
+        b2_tool, bucket_name, '', encryption=SSE_B2_AES, expected_encryption=SSE_B2_AES
+    )
+
+
+def sync_copy_helper(
+    b2_tool, bucket_name, folder_in_bucket, encryption=None, expected_encryption=SSE_NONE
+):
     file_to_upload = 'README.md'
 
     b2_sync_point = 'b2:%s' % bucket_name
@@ -862,22 +948,45 @@ def sync_copy_helper(b2_tool, bucket_name, folder_in_bucket):
 
     # Put a couple files in B2
     b2_tool.should_succeed(
-        ['upload-file', '--noProgress', bucket_name, file_to_upload, b2_file_prefix + 'a']
+        [
+            'upload-file', '--noProgress', '--destinationServerSideEncryption', 'SSE-B2',
+            bucket_name, file_to_upload, b2_file_prefix + 'a'
+        ]
     )
     b2_tool.should_succeed(
         ['upload-file', '--noProgress', bucket_name, file_to_upload, b2_file_prefix + 'b']
     )
 
     # Sync all the files
-    b2_tool.should_succeed(['sync', '--noProgress', b2_sync_point, other_b2_sync_point])
+    if encryption is None:
+        b2_tool.should_succeed(['sync', '--noProgress', b2_sync_point, other_b2_sync_point])
+    elif encryption == SSE_NONE:
+        b2_tool.should_succeed(
+            [
+                'sync', '--noProgress', '--destinationServerSideEncryption', 'none', b2_sync_point,
+                other_b2_sync_point
+            ]
+        )
+    elif encryption == SSE_B2_AES:
+        b2_tool.should_succeed(
+            [
+                'sync', '--noProgress', '--destinationServerSideEncryption', 'SSE-B2',
+                b2_sync_point, other_b2_sync_point
+            ]
+        )
+    else:
+        raise ValueError(expected_encryption)
 
     file_versions = b2_tool.list_file_versions(other_bucket_name)
+
+    expected_encryption_str = encryption_summary(expected_encryption.as_value_dict())
+
     should_equal(
         [
-            '+ ' + b2_file_prefix + 'a',
-            '+ ' + b2_file_prefix + 'b',
+            ('+ ' + b2_file_prefix + 'a', expected_encryption_str),
+            ('+ ' + b2_file_prefix + 'b', expected_encryption_str),
         ],
-        file_version_summary(file_versions),
+        file_version_summary_with_encryption(file_versions),
     )
 
 
@@ -907,6 +1016,112 @@ def sync_long_path_test(b2_tool, bucket_name):
         should_equal(['+ ' + long_path], file_version_summary(file_versions))
 
 
+def default_sse_b2_test(b2_tool, bucket_name):
+    # Set default encryption via update-bucket
+    bucket_info = b2_tool.should_succeed_json(['get-bucket', bucket_name])
+    bucket_default_sse = {'mode': 'none'}
+    should_equal(bucket_default_sse, bucket_info['defaultServerSideEncryption'])
+
+    bucket_info = b2_tool.should_succeed_json(
+        ['update-bucket', '--defaultServerSideEncryption=SSE-B2', bucket_name, 'allPublic']
+    )
+    bucket_default_sse = {
+        'isClientAuthorizedToRead': True,
+        'value': {
+            'algorithm': 'AES256',
+            'mode': 'SSE-B2',
+        }
+    }
+    should_equal(bucket_default_sse, bucket_info['defaultServerSideEncryption'])
+
+    bucket_info = b2_tool.should_succeed_json(['get-bucket', bucket_name])
+    bucket_default_sse = {
+        'algorithm': 'AES256',
+        'mode': 'SSE-B2',
+    }
+    should_equal(bucket_default_sse, bucket_info['defaultServerSideEncryption'])
+
+    # Set default encryption via create-bucket
+    second_bucket_name = b2_tool.bucket_name_prefix + '-' + random_hex(8)
+    b2_tool.should_succeed(
+        ['create-bucket', '--defaultServerSideEncryption=SSE-B2', second_bucket_name, 'allPublic']
+    )
+    second_bucket_info = b2_tool.should_succeed_json(['get-bucket', second_bucket_name])
+    second_bucket_default_sse = {
+        'algorithm': 'AES256',
+        'mode': 'SSE-B2',
+    }
+    should_equal(second_bucket_default_sse, second_bucket_info['defaultServerSideEncryption'])
+
+
+def sse_b2_test(b2_tool, bucket_name):
+    file_to_upload = 'README.md'
+
+    b2_tool.should_succeed(
+        [
+            'upload-file', '--destinationServerSideEncryption=SSE-B2', '--noProgress', '--quiet',
+            bucket_name, file_to_upload, 'encrypted'
+        ]
+    )
+    b2_tool.should_succeed(
+        ['upload-file', '--noProgress', '--quiet', bucket_name, file_to_upload, 'not_encrypted']
+    )
+
+    b2_tool.should_succeed(
+        ['download-file-by-name', '--noProgress', bucket_name, 'encrypted', os.devnull]
+    )
+    b2_tool.should_succeed(
+        ['download-file-by-name', '--noProgress', bucket_name, 'not_encrypted', os.devnull]
+    )
+
+    list_of_files = b2_tool.should_succeed_json(['ls', '--json', '--recursive', bucket_name])
+    should_equal(
+        [{
+            'algorithm': 'AES256',
+            'mode': 'SSE-B2'
+        }, {
+            'mode': 'none'
+        }], [f['serverSideEncryption'] for f in list_of_files]
+    )
+
+    encrypted_version = list_of_files[0]
+    file_info = b2_tool.should_succeed_json(['get-file-info', encrypted_version['fileId']])
+    should_equal({'algorithm': 'AES256', 'mode': 'SSE-B2'}, file_info['serverSideEncryption'])
+    not_encrypted_version = list_of_files[1]
+    file_info = b2_tool.should_succeed_json(['get-file-info', not_encrypted_version['fileId']])
+    #should_equal({'mode': 'none'}, file_info['serverSideEncryption'])  # v1 B2Api.get_file_info is a legacy interface which returns raw server response
+
+    b2_tool.should_succeed(
+        [
+            'copy-file-by-id', '--destinationServerSideEncryption=SSE-B2',
+            encrypted_version['fileId'], bucket_name, 'copied_encrypted'
+        ]
+    )
+    b2_tool.should_succeed(
+        ['copy-file-by-id', not_encrypted_version['fileId'], bucket_name, 'copied_not_encrypted']
+    )
+
+    list_of_files = b2_tool.should_succeed_json(['ls', '--json', '--recursive', bucket_name])
+    should_equal(
+        [{
+            'algorithm': 'AES256',
+            'mode': 'SSE-B2'
+        }, {
+            'mode': 'none'
+        }] * 2, [f['serverSideEncryption'] for f in list_of_files]
+    )
+
+    copied_encrypted_version = list_of_files[2]
+    file_info = b2_tool.should_succeed_json(['get-file-info', copied_encrypted_version['fileId']])
+    should_equal({'algorithm': 'AES256', 'mode': 'SSE-B2'}, file_info['serverSideEncryption'])
+
+    copied_not_encrypted_version = list_of_files[3]
+    file_info = b2_tool.should_succeed_json(
+        ['get-file-info', copied_not_encrypted_version['fileId']]
+    )
+    #should_equal({'mode': 'none'}, file_info['serverSideEncryption'])  # v1 B2Api.get_file_info is a legacy interface which returns raw server response
+
+
 def main(bucket_name_prefix):
     test_map = {
         'account': account_test,
@@ -915,11 +1130,16 @@ def main(bucket_name_prefix):
         'sync_down': sync_down_test,
         'sync_down_no_prefix': sync_down_test_no_prefix,
         'sync_up': sync_up_test,
+        'sync_up_sse_b2': sync_up_sse_b2_test,
         'sync_up_no_prefix': sync_up_test_no_prefix,
         'sync_long_path': sync_long_path_test,
         'sync_copy': sync_copy_test,
-        'sync_copy_no_prefix': sync_copy_test_no_prefix,
+        'sync_copy_test_no_prefix_default_encryption': sync_copy_test_no_prefix_default_encryption,
+        #'sync_copy_test_no_prefix_no_encryption': sync_copy_test_no_prefix_no_encryption, # not supported by the server
+        'sync_copy_test_no_prefix_sse_b2': sync_copy_test_no_prefix_sse_b2,
         'download': download_test,
+        'default_sse_b2': default_sse_b2_test,
+        'sse_b2': sse_b2_test,
     }
 
     args = parse_args(tests=sorted(test_map))
