@@ -290,6 +290,12 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "isClientAuthorizedToRead": true,
+                "value": {{
+                    "mode": "none"
+                }}
+            }},
             "lifecycleRules": [],
             "options": [],
             "revision": 2
@@ -311,6 +317,75 @@ class TestConsoleTool(TestBase):
 
         self._run_command(['delete-bucket', 'your-bucket'], expected_stdout, '', 0)
 
+    def test_encrypted_buckets(self):
+        self._authorize_account()
+
+        # Make two encrypted buckets
+        self._run_command(['create-bucket', 'my-bucket', 'allPrivate'], 'bucket_0\n', '', 0)
+        self._run_command(
+            ['create-bucket', '--defaultServerSideEncryption=SSE-B2', 'your-bucket', 'allPrivate'],
+            'bucket_1\n', '', 0
+        )
+
+        # Update the one without encryption
+        expected_stdout = '''
+        {{
+            "accountId": "{account_id}",
+            "bucketId": "bucket_0",
+            "bucketInfo": {{}},
+            "bucketName": "my-bucket",
+            "bucketType": "allPublic",
+            "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "isClientAuthorizedToRead": true,
+                "value": {{
+                    "algorithm": "AES256",
+                    "mode": "SSE-B2"
+                }}
+            }},
+            "lifecycleRules": [],
+            "options": [],
+            "revision": 2
+        }}
+        '''
+
+        self._run_command(
+            ['update-bucket', '--defaultServerSideEncryption=SSE-B2', 'my-bucket', 'allPublic'],
+            expected_stdout, '', 0
+        )
+
+        # Update the one with encryption
+        expected_stdout = '''
+        {{
+            "accountId": "{account_id}",
+            "bucketId": "bucket_1",
+            "bucketInfo": {{}},
+            "bucketName": "your-bucket",
+            "bucketType": "allPrivate",
+            "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "isClientAuthorizedToRead": true,
+                "value": {{
+                    "algorithm": "AES256",
+                    "mode": "SSE-B2"
+                }}
+            }},
+            "lifecycleRules": [],
+            "options": [],
+            "revision": 2
+        }}
+        '''
+
+        self._run_command(['update-bucket', 'your-bucket', 'allPrivate'], expected_stdout, '', 0)
+
+        # Make sure they are there
+        expected_stdout = '''
+        bucket_0  allPublic   my-bucket
+        bucket_1  allPrivate  your-bucket
+        '''
+
+        self._run_command(['list-buckets'], expected_stdout, '', 0)
+
     def test_keys(self):
         self._authorize_account()
 
@@ -354,7 +429,10 @@ class TestConsoleTool(TestBase):
             0,
         )
         self._run_command(
-            ['create-key', '--bucket', 'my-bucket-a', 'goodKeyName-Two', capabilities_with_commas],
+            [
+                'create-key', '--bucket', 'my-bucket-a', 'goodKeyName-Two',
+                capabilities_with_commas + ',readBucketEncryption'
+            ],
             'appKeyId1 appKey1\n',
             '',
             0,
@@ -384,7 +462,7 @@ class TestConsoleTool(TestBase):
 
         expected_list_keys_out_long = """
             appKeyId0   goodKeyName-One        -                      -            -          ''   readFiles,listBuckets
-            appKeyId1   goodKeyName-Two        my-bucket-a            -            -          ''   readFiles,listBuckets
+            appKeyId1   goodKeyName-Two        my-bucket-a            -            -          ''   readFiles,listBuckets,readBucketEncryption
             appKeyId2   goodKeyName-Three      id=bucket_1            -            -          ''   readFiles,listBuckets
             """
 
@@ -409,6 +487,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket-a",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": null
+            }},
             "lifecycleRules": [],
             "options": [],
             "revision": 1
@@ -438,6 +519,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket-a",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "lifecycleRules": [],
             "options": [],
             "revision": 1
@@ -467,6 +551,12 @@ class TestConsoleTool(TestBase):
                 "bucketName": "my-bucket",
                 "bucketType": "allPrivate",
                 "corsRules": [],
+                "defaultServerSideEncryption": {{
+                    "isClientAuthorizedToRead": true,
+                    "value": {{
+                        "mode": "none"
+                    }}
+                }},
                 "lifecycleRules": [],
                 "options": [],
                 "revision": 2
@@ -525,6 +615,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "1500111222000"
                 }},
                 "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "size": 11,
                 "uploadTimestamp": 5000
             }}
@@ -550,6 +643,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "1500111222000"
                 }},
                 "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "uploadTimestamp": 5000
             }}
             '''
@@ -593,6 +689,9 @@ class TestConsoleTool(TestBase):
                 "fileId": "9998",
                 "fileInfo": {{}},
                 "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "size": 0,
                 "uploadTimestamp": 5001
             }}
@@ -609,6 +708,9 @@ class TestConsoleTool(TestBase):
                     "fileId": "9998",
                     "fileInfo": {{}},
                     "fileName": "file1.txt",
+                    "serverSideEncryption": {{
+                        "mode": "none"
+                    }},
                     "size": 0,
                     "uploadTimestamp": 5001
                 }},
@@ -621,6 +723,188 @@ class TestConsoleTool(TestBase):
                         "src_last_modified_millis": "%s"
                     }},
                     "fileName": "file1.txt",
+                    "serverSideEncryption": {{
+                        "mode": "none"
+                    }},
+                    "size": 11,
+                    "uploadTimestamp": 5000
+                }}
+            ]
+            ''' % (mod_time_str,)
+
+            self._run_command(['ls', '--json', '--versions', 'my-bucket'], expected_stdout, '', 0)
+
+            # List the file names
+            expected_stdout = '''
+            []
+            '''
+
+            self._run_command(['ls', '--json', 'my-bucket'], expected_stdout, '', 0)
+
+            # Delete one file version, passing the name in
+            expected_stdout = '''
+            {{
+                "action": "delete",
+                "fileId": "9998",
+                "fileName": "file1.txt"
+            }}
+            '''
+
+            self._run_command(['delete-file-version', 'file1.txt', '9998'], expected_stdout, '', 0)
+
+            # Delete one file version, not passing the name in
+            expected_stdout = '''
+            {{
+                "action": "delete",
+                "fileId": "9999",
+                "fileName": "file1.txt"
+            }}
+            '''
+
+            self._run_command(['delete-file-version', '9999'], expected_stdout, '', 0)
+
+    def test_files_encrypted(self):
+
+        self._authorize_account()
+        self._run_command(['create-bucket', 'my-bucket', 'allPublic'], 'bucket_0\n', '', 0)
+
+        with TempDir() as temp_dir:
+            local_file1 = self._make_local_file(temp_dir, 'file1.txt')
+            # For this test, use a mod time without millis.  My mac truncates
+            # millis and just leaves seconds.
+            mod_time = 1500111222
+            os.utime(local_file1, (mod_time, mod_time))
+            self.assertEqual(1500111222, os.path.getmtime(local_file1))
+
+            # Upload a file
+            expected_stdout = '''
+            URL by file name: http://download.example.com/file/my-bucket/file1.txt
+            URL by fileId: http://download.example.com/b2api/vx/b2_download_file_by_id?fileId=9999
+            {{
+                "action": "upload",
+                "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
+                "contentType": "b2/x-auto",
+                "fileId": "9999",
+                "fileInfo": {{
+                    "src_last_modified_millis": "1500111222000"
+                }},
+                "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "algorithm": "AES256",
+                    "mode": "SSE-B2"
+                }},
+                "size": 11,
+                "uploadTimestamp": 5000
+            }}
+            '''
+
+            self._run_command(
+                [
+                    'upload-file', '--noProgress', '--destinationServerSideEncryption=SSE-B2',
+                    'my-bucket', local_file1, 'file1.txt'
+                ], expected_stdout, '', 0, None, True
+            )
+
+            # Get file info
+            mod_time_str = str(file_mod_time_millis(local_file1))
+            expected_stdout = '''
+            {{
+                "accountId": "{account_id}",
+                "action": "upload",
+                "bucketId": "bucket_0",
+                "contentLength": 11,
+                "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
+                "contentType": "b2/x-auto",
+                "fileId": "9999",
+                "fileInfo": {{
+                    "src_last_modified_millis": "1500111222000"
+                }},
+                "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "algorithm": "AES256",
+                    "mode": "SSE-B2"
+                }},
+                "uploadTimestamp": 5000
+            }}
+            '''
+
+            self._run_command(['get-file-info', '9999'], expected_stdout, '', 0)
+
+            # Download by name
+            local_download1 = os.path.join(temp_dir, 'download1.txt')
+            expected_stdout = '''
+            File name:    file1.txt
+            File id:      9999
+            File size:    11
+            Content type: b2/x-auto
+            Content sha1: 2aae6c35c94fcfb415dbe95f408b9ce91ee846ed
+            INFO src_last_modified_millis: 1500111222000
+            checksum matches
+            '''
+
+            self._run_command(
+                [
+                    'download-file-by-name', '--noProgress', 'my-bucket', 'file1.txt',
+                    local_download1
+                ], expected_stdout, '', 0
+            )
+            self.assertEqual(b'hello world', self._read_file(local_download1))
+            self.assertEqual(mod_time, int(round(os.path.getmtime(local_download1))))
+
+            # Download file by ID.  (Same expected output as downloading by name)
+            local_download2 = os.path.join(temp_dir, 'download2.txt')
+            self._run_command(
+                ['download-file-by-id', '--noProgress', '9999', local_download2], expected_stdout,
+                '', 0
+            )
+            self.assertEqual(b'hello world', self._read_file(local_download2))
+
+            # Hide the file
+            expected_stdout = '''
+            {{
+                "action": "hide",
+                "contentSha1": "none",
+                "fileId": "9998",
+                "fileInfo": {{}},
+                "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
+                "size": 0,
+                "uploadTimestamp": 5001
+            }}
+            '''
+
+            self._run_command(['hide-file', 'my-bucket', 'file1.txt'], expected_stdout, '', 0)
+
+            # List the file versions
+            expected_stdout = '''
+            [
+                {{
+                    "action": "hide",
+                    "contentSha1": "none",
+                    "fileId": "9998",
+                    "fileInfo": {{}},
+                    "fileName": "file1.txt",
+                    "serverSideEncryption": {{
+                        "mode": "none"
+                    }},
+                    "size": 0,
+                    "uploadTimestamp": 5001
+                }},
+                {{
+                    "action": "upload",
+                    "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
+                    "contentType": "b2/x-auto",
+                    "fileId": "9999",
+                    "fileInfo": {{
+                        "src_last_modified_millis": "%s"
+                    }},
+                    "fileName": "file1.txt",
+                    "serverSideEncryption": {{
+                        "algorithm": "AES256",
+                        "mode": "SSE-B2"
+                    }},
                     "size": 11,
                     "uploadTimestamp": 5000
                 }}
@@ -683,6 +967,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "1500111222000"
                 }},
                 "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "size": 11,
                 "uploadTimestamp": 5000
             }}
@@ -707,6 +994,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "1500111222000"
                 }},
                 "fileName": "file1_copy.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "uploadTimestamp": 5001
             }}
             '''
@@ -728,6 +1018,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "1500111222000"
                 }},
                 "fileName": "file1_copy.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "uploadTimestamp": 5002
             }}
             '''
@@ -782,6 +1075,9 @@ class TestConsoleTool(TestBase):
                     "a": "b"
                 }},
                 "fileName": "file1_copy.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "uploadTimestamp": 5003
             }}
             '''
@@ -827,6 +1123,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "1500111222000"
                 }},
                 "fileName": "file1_copy.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "uploadTimestamp": 5004
             }}
             '''
@@ -883,10 +1182,12 @@ class TestConsoleTool(TestBase):
         large_file_upload_state = mock.MagicMock()
         large_file_upload_state.has_error.return_value = False
         bucket.api.services.upload_manager._upload_part(
-            bucket.id_, file.file_id, UploadSourceBytes(content), 1, large_file_upload_state
+            bucket.id_, file.file_id, UploadSourceBytes(content), 1, large_file_upload_state, None,
+            None
         )
         bucket.api.services.upload_manager._upload_part(
-            bucket.id_, file.file_id, UploadSourceBytes(content), 3, large_file_upload_state
+            bucket.id_, file.file_id, UploadSourceBytes(content), 3, large_file_upload_state, None,
+            None
         )
         expected_stdout = '''
             1         11  2aae6c35c94fcfb415dbe95f408b9ce91ee846ed
@@ -944,6 +1245,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "%s"
                 }},
                 "fileName": "test.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "size": 600,
                 "uploadTimestamp": 5000
             }}
@@ -956,6 +1260,46 @@ class TestConsoleTool(TestBase):
                 ], expected_stdout, '', 0, None, True
             )
 
+    def test_upload_large_file_encrypted(self):
+        self._authorize_account()
+        self._run_command(['create-bucket', 'my-bucket', 'allPublic'], 'bucket_0\n', '', 0)
+        min_part_size = self.account_info.get_minimum_part_size()
+        file_size = min_part_size * 3
+
+        with TempDir() as temp_dir:
+            file_path = os.path.join(temp_dir, 'test.txt')
+            text = '*' * file_size
+            with open(file_path, 'wb') as f:
+                f.write(text.encode('utf-8'))
+            mod_time_str = str(file_mod_time_millis(file_path))
+            expected_stdout = '''
+            URL by file name: http://download.example.com/file/my-bucket/test.txt
+            URL by fileId: http://download.example.com/b2api/vx/b2_download_file_by_id?fileId=9999
+            {{
+                "action": "upload",
+                "contentSha1": "none",
+                "contentType": "b2/x-auto",
+                "fileId": "9999",
+                "fileInfo": {{
+                    "src_last_modified_millis": "%s"
+                }},
+                "fileName": "test.txt",
+                "serverSideEncryption": {{
+                    "algorithm": "AES256",
+                    "mode": "SSE-B2"
+                }},
+                "size": 600,
+                "uploadTimestamp": 5000
+            }}
+            ''' % (mod_time_str,)
+
+            self._run_command(
+                [
+                    'upload-file', '--noProgress', '--destinationServerSideEncryption=SSE-B2',
+                    '--threads', '5', 'my-bucket', file_path, 'test.txt'
+                ], expected_stdout, '', 0, None, True
+            )
+
     def test_get_account_info(self):
         self._authorize_account()
         expected_stdout = '''
@@ -972,6 +1316,8 @@ class TestConsoleTool(TestBase):
                     "listBuckets",
                     "writeBuckets",
                     "deleteBuckets",
+                    "readBucketEncryption",
+                    "writeBucketEncryption",
                     "listFiles",
                     "readFiles",
                     "shareFiles",
@@ -998,6 +1344,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "lifecycleRules": [],
             "options": [],
             "revision": 1
@@ -1016,6 +1365,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "fileCount": 0,
             "lifecycleRules": [],
             "options": [],
@@ -1044,6 +1396,9 @@ class TestConsoleTool(TestBase):
                     "src_last_modified_millis": "%s"
                 }},
                 "fileName": "file1.txt",
+                "serverSideEncryption": {{
+                    "mode": "none"
+                }},
                 "size": 11,
                 "uploadTimestamp": 5000
             }}
@@ -1062,6 +1417,9 @@ class TestConsoleTool(TestBase):
                 "bucketName": "my-bucket",
                 "bucketType": "allPublic",
                 "corsRules": [],
+                "defaultServerSideEncryption": {{
+                    "mode": "none"
+                }},
                 "fileCount": 1,
                 "lifecycleRules": [],
                 "options": [],
@@ -1097,6 +1455,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "fileCount": 10,
             "lifecycleRules": [],
             "options": [],
@@ -1143,6 +1504,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "fileCount": 20,
             "lifecycleRules": [],
             "options": [],
@@ -1183,6 +1547,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "fileCount": 10,
             "lifecycleRules": [],
             "options": [],
@@ -1243,6 +1610,9 @@ class TestConsoleTool(TestBase):
             "bucketName": "my-bucket",
             "bucketType": "allPublic",
             "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "mode": "none"
+            }},
             "fileCount": 29,
             "lifecycleRules": [],
             "options": [],
@@ -1252,6 +1622,35 @@ class TestConsoleTool(TestBase):
         '''
         self._run_command(['get-bucket', '--showSize', 'my-bucket'], expected_stdout, '', 0)
 
+    def test_get_bucket_encrypted(self):
+        self._authorize_account()
+        self._run_command(
+            [
+                'create-bucket', '--defaultServerSideEncryption=SSE-B2',
+                '--defaultServerSideEncryptionAlgorithm=AES256', 'my-bucket', 'allPublic'
+            ], 'bucket_0\n', '', 0
+        )
+        expected_stdout = '''
+        {{
+            "accountId": "{account_id}",
+            "bucketId": "bucket_0",
+            "bucketInfo": {{}},
+            "bucketName": "my-bucket",
+            "bucketType": "allPublic",
+            "corsRules": [],
+            "defaultServerSideEncryption": {{
+                "algorithm": "AES256",
+                "mode": "SSE-B2"
+            }},
+            "fileCount": 0,
+            "lifecycleRules": [],
+            "options": [],
+            "revision": 1,
+            "totalSize": 0
+        }}
+        '''
+        self._run_command(['get-bucket', '--showSize', 'my-bucket'], expected_stdout, '', 0)
+
     def test_sync(self):
         self._authorize_account()
         self._create_my_bucket()
@@ -1327,6 +1726,9 @@ class TestConsoleTool(TestBase):
                         "src_last_modified_millis": "%d"
                     }},
                     "fileName": "test-dry-run.txt",
+                    "serverSideEncryption": {{
+                        "mode": "none"
+                    }},
                     "size": 11,
                     "uploadTimestamp": 5000
                 }}
